//=============================================================================
// Project:    SLP - An Interpreter for Super Logic Programs / Static Semantics
// Module:     gneg.h
// Purpose:    Ground default negations (conditions in conditional facts)
// Last Change:04.03.2002
// Language:   C++
// Author:     Stefan Brass
// Email:      Stefan.Brass@informatik.uni-giessen.de, sbrass@sis.pitt.edu
// Address:    Universitaet Giessen, Arndtstr. 2, 35392 Giessen, Germany
// Copyright:  (c) 2001-2002 by Stefan Brass
// Licence:    Distributed under the conditions of the GNU Public Licence.
//=============================================================================

//=============================================================================
// Include File Frame:
//=============================================================================

#ifndef	GNEG_INCLUDED
#define GNEG_INCLUDED

//=============================================================================
// Used Types and Macros:
//=============================================================================

#ifndef VER_INCLUDED
#include "ver.h"
#endif

#ifndef BOOL_INCLUDED
#include "bool.h"
#endif

#ifndef CHECK_INCLUDED
#include "check.h"
#endif

#ifndef LIT_INCLUDED
#include "lit.h"
#endif

#ifndef GLIT_INCLUDED
#include "glit.h"
#endif

//=============================================================================
// Private Constants:
//=============================================================================

//-----------------------------------------------------------------------------
// GNEG_MAGIC: Magic number (identifies objects of gneg_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	GNEG_MAGIC =	0x474E4500L;	// 'GNE\0'
#endif

//-----------------------------------------------------------------------------
// GNEG_L_MAGIC: Magic number (identifies objects of gneg_l_c for debugging).
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
static const long	GNEG_L_MAGIC =	0x474E4C00L;	// 'GNL\0'
#endif

//=============================================================================
// Class for Ground Default Negations (Conditions in Conditional Facts):
//=============================================================================

class gneg_c {

//-----------------------------------------------------------------------------
// Static Member Functions (Class Methods):
//-----------------------------------------------------------------------------

	public:

	// store: Create or find ground default negation:
	static gneg_c* store(glit_t *lits);

	// store: Create or find ground default negation:
	static gneg_c* store(lit_t *lits);

	// Print all currently existing default negations:
	static void print_all(void);

	// Start duplicate check for a new set of ground default negations:
	static void	dup_ck_begin(void);

	// Close current set of ground default negations:
	static inline void	dup_ck_end(void)
		{ CHECK(SetIsOpen,
			"gneg_c::dup_ck_end: Duplicate check set not open");
		  SetIsOpen = BOOL_FALSE; }

	// Number remaining "critical" default negations (in residual program):
	static int number_crit_neg(void);

	// Get critical default negation with given number:
	static inline gneg_c*	get_crit_neg(int no)
		{ CHECK(NumCrit >= 0, "gneg_c::get_crit_neg: not initialized");
		  CHECK(no >= 0, "gneg_c::get_crit_neg: no is negative");
		  CHECK(no <= NumCrit, "gneg_c::get_crit_neg: no too large");
		  return CritNegs[no]; }

//-----------------------------------------------------------------------------
// Member Functions (Object Methods):
//-----------------------------------------------------------------------------

	public:

	// Get the list of facts inside the default negation:
	inline class gneg_l_c* get_elems(void) const
		{ CHECK(valid(), "gneg_c::get_elems: Bad object");
		  return(Lits); }

	// Get list of occurrences of this default negation in cond facts:
	inline class cf_cn_c* get_occur(void) const
		{ CHECK(valid(), "gneg_c::get_occur: Bad object");
		  return(Occur); }

	// Set the list of occurrences of this default negation in cond facts:
	inline void set_occur(class cf_cn_c *occur)
		{ CHECK(valid(), "gneg_c::set_occur: Bad object");
		  Occur = occur; }

	// This object is member of the current set. Is it already contained?
	inline bool_t	dup_ck_is_new(void)
		{ CHECK(valid(), "gneg_c::dup_ck_is_new: Invalid Object");
		  CHECK(SetIsOpen, "gneg_c::dup_ck_is_new: No set open");
		  if(MemberOf == CurrentSet)
			  return(BOOL_FALSE);
		  MemberOf = CurrentSet;
		  return(BOOL_TRUE); }

	// Check whether this negation contains more than one literal:
	bool_t	is_conjunctive(void) const;

	// Print this default negation:
	void print(void) const;

	// Number of this default negation (must be "critical"):
	inline int	get_no(void)
		{ CHECK(valid(), "gneg_c::no: Invalid Object");
		  CHECK(No >= 0, "gneg_c::no: Not critical");
		  CHECK(No < VER_MAX_CRIT_NEG, "gneg_c::no: Number invalid");
		  return(No); }

//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "GNEG_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Static Data Members (Class Attributes):
//-----------------------------------------------------------------------------

	private:

	// List of all existing ground default negations (gneg_c objects):
	static gneg_c* Negs;

	// Pointer to last element of linked list of all gneg_c objects:
	static gneg_c* LastNeg;

	// Unique number of current set of default negations (for dup_ck):
	static long CurrentSet;

	// Is currently a set of default negations open (for dup_ck)?
	static bool_t SetIsOpen; // Only needed for assertions (CHECK)

	// Unique number of current comparison of default negations:
	static long NegComparID;

	// Number of critical default negations (in residual program):
	static int NumCrit;

	// Array of critical default negations (in sequence of gneg_c::no):
	static gneg_c *CritNegs[VER_MAX_CRIT_NEG];

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Link to next default negation:
	gneg_c *Next;

	// List of elements of the conjunction inside this default negation:
	class gneg_l_c *Lits;

	// Last element of the conjunction list so far:
	class gneg_l_c *LastLit;

	// Number of ground literals inside this default negation:
	int Length;

	// Overlap counter (number of common literals for current comparison):
	int Overlap;

	// Number of the comparison (NegComparID) when Overlap was last set:
	long LastSet;

	// MemberOf: Number of last set this object appeared in (for dup_ck):
	long MemberOf;

	// List of conditional fact body elements that are this default neg:
	class cf_cn_c *Occur;

	// Number of this default negation (-1 if not critical):
	int No;

//-----------------------------------------------------------------------------
// Auxillary Member Functions (Private Object Methods):
//-----------------------------------------------------------------------------

	private:

	// Constructor:
	gneg_c(void);

	// Destructor (should never be called):
	~gneg_c(void);

	// Add ground literal to  the conjunction inside the default negation:
	void add_elem(class gneg_l_c *elem);

	// Is this the set of ground literals currently searched for?
	bool_t equal(long comparison, int length);

	// Get pointer to next default negation in the list of all:
	inline gneg_c* get_next(void) const
		{ CHECK(valid(), "gneg_c::next: Bad object");
		  return(Next); }

	// Set the Next pointer for the list of all default negations:
	inline void set_next(gneg_c *next)
		{ CHECK(valid(), "gneg_c::set_next: Bad object");
		  Next = next; }

	// Set the critical negation number:
	inline void set_no(int no)
		{ CHECK(valid(), "gneg_c::set_no: Bad object");
		  CHECK(no >= 0, "gneg_c::set_no: no is negative");
		  CHECK(no < VER_MAX_CRIT_NEG,
				"gneg_c::set_no: no >= VER_MAX_CRIT_NEG");
		  CHECK(Occur!=0, "gneg_c::set_no: negation is not critical");
		  No = no; }

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef gneg_c *gneg_t;

//=============================================================================
// Class for Elements of the Conjunction Inside the Default Negation (Facts):
//=============================================================================

class gneg_l_c {

	public:

//-----------------------------------------------------------------------------
// Methods:
//-----------------------------------------------------------------------------

	// Constructor:
	gneg_l_c(glit_t lit, gneg_c *neg);

	// Destructor:
	~gneg_l_c(void);

	// Return the default negation to which this fact belongs:
	inline gneg_c *get_neg(void) const
		{ CHECK(valid(), "gneg_l_c::cf: Bad object");
		  return(Neg); }

	// Return the ground literal of this default negation element:
	inline glit_t get_lit(void) const
		{ CHECK(valid(), "gneg_l_c::get_lit: Bad object");
		  return(Lit); }

	// Next occurrence of the same ground literal in a default negation:
	inline gneg_l_c *get_next_lit_occur(void) const
		{ CHECK(valid(), "gneg_l_c::get_next_lit_occur: Bad object");
		  return(NextLitOccur); }

	// Set link to the next occurrence of the same ground literal:
	inline void set_next_lit_occur(gneg_l_c *next)
		{ CHECK(valid(), "gneg_l_c::set_next_lit_occur: Bad object");
		  CHECK(next == 0 || next->valid(),
			"gneg_l_c::set_next_lit_occur: next is invalid");
		  NextLitOccur = next; }

	// Return the next element of the same default negation:
	inline gneg_l_c *get_next_neg_elem(void) const
		{ CHECK(valid(), "gneg_l_c::get_next_neg_elem: Bad object");
		  return(NextNegElem); }

	// Append an element to the list of literals in a default negation:
	void set_next_neg_elem(gneg_l_c *elem)
		{ CHECK(valid(), "gneg_l_c::set_next_neg_elem: Bad object");
		  CHECK(elem != 0, "gneg_l_c::set_next_neg_elem: hd is NIL");
		  CHECK(elem->valid(),
			"gneg_l_c::set_next_neg_elem: elem is invalid");
		  CHECK(NextNegElem == 0,
			"gneg_l_c::set_next_neg_elem: NextNegElem aready set");
		  NextNegElem = elem; }


//-----------------------------------------------------------------------------
// Debugging Support:
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
	// Integrity Check:
	public:
	bool_t	valid(void) const;

	// Magic number (identifies objects of this class for debugging).
	private:
	long	Magic;		// Must be "GNEG_L_MAGIC"
#endif

//-----------------------------------------------------------------------------
// Data Members (Object Attributes):
//-----------------------------------------------------------------------------

	private:

	// Reference back to Default Negation:
	gneg_c	*Neg;

	// Fact in this default negation element:
	glit_t	Lit;

	// Pointer to next element of the same default negation:
	gneg_l_c *NextNegElem;

	// Next occurrence of the same ground literal in a default negation:
	gneg_l_c *NextLitOccur;

//-----------------------------------------------------------------------------
// End of Class:
//-----------------------------------------------------------------------------

};

//-----------------------------------------------------------------------------
// Declaration of Pointer Type:
//-----------------------------------------------------------------------------

typedef gneg_l_c *gneg_l_t;

//=============================================================================
// End of Include File:
//=============================================================================

#endif

